///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoBTreeNode.cc
// -----------------
// Cego btree node class implementation
//     
// Design and Implementation by Bjoern Lemke               
//     
// (C)opyright 2000-2017 Bjoern Lemke
//
// IMPLEMENTATION MODULE
//
// Class: CegoBTreeNode
// 
// Description: This class handles operations on btree elements.
//              These elements either can be nodes or leafs and depending on the type, the corresponding methods are available.
//
//
// Status: CLEAN
//
///////////////////////////////////////////////////////////////////////////////

// LFC INCLUDES
#include <lfcbase/Exception.h>

// CEGO INCLUDES
#include "CegoBTreeNode.h"
#include "CegoDataPointer.h"
#include "CegoTupleState.h"
#include "CegoQueryHelper.h"

// POSIX INCLUDES
#include <string.h>
#include <stdlib.h>

// must match the value of method CegoDataPointer.getEncodingLength 
#define DPENCODINGLENGTH CegoDataPointer::getEncodingLength()

char __allowDuplicateNull = 0;


CegoBTreeNode::CegoBTreeNode()
{
    _pI = 0;
    _len = 0;
    _keyLen = 0;
    _maxEntries = 0;
    _pageId = 0;
    _nextPageId = 0;
    _pNextChild = 0;
    _nt = CegoBTreeNode::UNDEFINED;
}

CegoBTreeNode::~CegoBTreeNode()
{
}

const CegoBTreeNode::NodeType CegoBTreeNode::getType() const
{
    return _nt;
}

void CegoBTreeNode::setPtr(void* p, int len)
{
    _pI = p;
    _len = len;
}

void* CegoBTreeNode::getPtr()
{
    return _pI;
}

int CegoBTreeNode::getLen() const
{
    return _len;
}


void CegoBTreeNode::setPageId(PageIdType pageId)
{
    _pageId = pageId;
}

PageIdType CegoBTreeNode::getPageId() const
{
    return _pageId;
}


void CegoBTreeNode::setNextPageId(PageIdType pageId)
{
    _nextPageId = pageId;
}

PageIdType CegoBTreeNode::getNextPageId() const
{
    return _nextPageId;
}

void CegoBTreeNode::initNode()
{
    if ( _pI )
    {
	(*(int*)(_pI)) = 0;
	/*
	int numEntry = 0;
	memcpy(_pI, &numEntry, sizeof(int));
	*/
	return;
    }
    throw Exception(EXLOC, Chain("Btree node not set up"));
}

int CegoBTreeNode::numEntries() const
{
    if ( _pI )
    {
	return (*(int*)(_pI));
	/*
	int n;
	memcpy(&n, _pI, sizeof(int));
	return n;
	*/
    }
    throw Exception(EXLOC, Chain("Btree node not set up"));
}

void CegoBTreeNode::incEntries()
{
    if ( _pI )
    {
	(*(int*)(_pI))++;
	/*
	int n;

	memcpy(&n, _pI, sizeof(int));
	n++;
	memcpy(_pI, &n, sizeof(int));
	*/
	return;
    }
    throw Exception(EXLOC, Chain("Btree node not set up"));
}

void CegoBTreeNode::decEntries()
{
    if ( _pI )
    {
	(*(int*)(_pI))--;

	/*
	int n;
	memcpy(&n, _pI, sizeof(int));
	n--;
	memcpy(_pI, &n, sizeof(int));
	*/
	return;
    }
    throw Exception(EXLOC, Chain("Btree node not set up"));
}

int CegoBTreeNode::getEntrySize() const
{
    if ( _nt == LEAF )
    {
	return _keyLen + DPENCODINGLENGTH;
    }
    else if ( _nt == NODE )
    {	
	return _keyLen + sizeof(PageIdType);
    }
    else
    {
	throw Exception(EXLOC, Chain("Unknown btree node type"));
    }
}

/* old stuff

void CegoBTreeNode::getChildPage(const CegoBTreeValue& iv, PageIdType& pageId)
{

    if ( _nt == LEAF )
	throw Exception(EXLOC, Chain("Invalid method for node type"));

    int ne = numEntries();

    if ( ne == 0 )
	throw Exception(EXLOC, "No child entries avaiable");

    int i=0;
    bool posFound=false;

    while ( i < ne && posFound == false)
    {
		
	char* pKey = (char*)((long long)_pI + sizeof(int)) + sizeof(PageIdType) + i * getEntrySize();
	CegoBTreeValue nv;
	nv.setPtr(pKey, _keyLen);

	if ( iv.isHigher(nv, _pSchema) )
	{	    
	    i++;
	}
	else
	{
	    posFound=true;	    
	}
    }

    char* p = (char*)((long long)_pI + sizeof(int)) + i * getEntrySize();
    
    memcpy(&pageId, p, sizeof(PageIdType));

    // cout << "Found child page " << pageId << endl;

    return;
}
*/

void CegoBTreeNode::getChildPage(const CegoBTreeValue& iv, PageIdType& pageId)
{

    if ( _nt == LEAF )
	throw Exception(EXLOC, Chain("Invalid method for node type"));

    int ne = numEntries();

    if ( ne == 0 )
	throw Exception(EXLOC, "No child entries avaiable");

    int lb=0;
    int rb=ne;

    int i;
    bool posFound=false;
    while ( posFound == false)
    {
		
	i = (lb+rb) / 2;

	char* pKey = (char*)((long long)_pI + sizeof(int)) + sizeof(PageIdType) + i * getEntrySize();
	CegoBTreeValue nv;
	nv.setPtr(pKey, _keyLen);

	// cout << "getChildPage : Comparing " << iv.toChain(_pSchema) << " and " << nv.toChain(_pSchema) << "i=" << i << "ne=" << ne << endl;
	
	if ( iv.isHigher(nv, _pSchema) )
	{	    
	    lb=i;
	}
	else
	{
	    rb=i;	    
	}
	
	if ( rb - lb <= 1 )
	{
	    if ( lb != rb )
	    {
		// we still have to check the left value
		char* pLeftKey = (char*)((long long)_pI + sizeof(int)) + sizeof(PageIdType) + lb * getEntrySize();
		CegoBTreeValue lv;
		lv.setPtr(pLeftKey, _keyLen);
		
		if ( iv.isHigher(lv, _pSchema) )
		{	    
		    i=rb;
		}
		else
		{
		    i=lb;
		}
	    }
	    else
	    {
		// evaluation can be skipped
		i=rb;
	    }

	    posFound=true;
	}      
    }

    // check pos
    /*
    int j=0;
    posFound=false;

    while ( j < ne && posFound == false)
    {
		
	char* pKey = (char*)((long long)_pI + sizeof(int)) + sizeof(PageIdType) + j * getEntrySize();
	CegoBTreeValue nv;
	nv.setPtr(pKey, _keyLen);

	if ( iv.isHigher(nv, _pSchema) )
	{	    
	    j++;
	}
	else
	{
	    posFound=true;	    
	}
    }

    if ( i != j )
    {
	cout << "POSITION MISMATCH" << endl;
	cout << "Value is " << iv.toChain(_pSchema) << endl;
	cout << "Offset is " <<  sizeof(int) + 2 * sizeof(int) + i * getEntrySize() << endl;
	cout << "Invalid position is " << i << endl;
	cout << "Valid position is " << j << endl;
	cout << "NumEntries are " << ne << endl;
	printNode();
	throw Exception(EXLOC, "MISMATCH TERMINATION");
    }
    */

    char* p = (char*)((long long)_pI + sizeof(int)) + i * getEntrySize();
    
    memcpy(&pageId, p, sizeof(PageIdType));

    // cout << "Found child page " << pageId << endl;

    return;
}


void CegoBTreeNode::setSchema(NodeType nt, ListT<CegoField>* pSchema, int keyLen)
{
    _pSchema = pSchema;
    _keyLen = keyLen;
    _nt = nt;

    if ( _nt == LEAF )
    {
	_maxEntries = ( _len - sizeof(int) )  / ( _keyLen + DPENCODINGLENGTH );
    }
    else if ( _nt == NODE )
    {	
	_maxEntries = ( _len - sizeof(int) - sizeof(PageIdType) ) / ( _keyLen + sizeof(PageIdType) );
    }

}

bool CegoBTreeNode::addValue(const CegoBTreeValue& iv, const CegoDataPointer& dp, const Chain& btreeName, bool isUnique, CegoObjectManager *pOM, int tabSetId, unsigned long long tid, bool& duplicateAdded)
{

    duplicateAdded = false;

    if ( isUnique && __allowDuplicateNull && iv.hasNull(*_pSchema))
	isUnique=false;
    
    if ( _nt == NODE )
      throw Exception(EXLOC, Chain("Invalid method for node type"));
    
    int ne = numEntries();
    
    if ( ne == _maxEntries )
    {
	CegoDataPointer dp1;
	if ( valueExists(iv, dp1) )
	{
	    // if we have detected duplicate on unique btree, we can throw exception
	    if ( isUnique && ! ( tid && checkForDeleted(pOM, dp1, tabSetId, tid) ) )
	    {		
		Chain msg = Chain("Duplicate key ") + iv.toChain(_pSchema) + Chain(" on unique btree ") + btreeName;
		throw Exception(EXLOC, msg);		    
	    }
	    
	    duplicateAdded=true;
	}

	return false;
    }

    if ( ne == 0 )
    {
	char* pKey = (char*)((long long)_pI + sizeof(int));
	memcpy(pKey, iv.getPtr(), _keyLen);
	char* pDP = (char*)((long long)pKey + _keyLen);
	dp.encode(pDP);
	
	incEntries();

	return true;	
    }

    bool isInserted = false;
        
    int lb=0;
    int rb=ne;

    int i = (lb+rb) / 2;
    
    bool posFound=false;

    while ( posFound == false )
    {
	// cout << "lb=" << lb << " rb=" << rb << endl;

	i = (lb+rb) / 2;

	// cout << "Checking position " << i << endl;

	CegoBTreeValue nv1;
	CegoBTreeValue nv2;

	CegoDataPointer dp1;
	CegoDataPointer dp2;

	nv1.setPtr((char*)((long long)_pI + sizeof(int)) + i * getEntrySize(), _keyLen);
	dp1.decode((char*)((long long)_pI + sizeof(int)) + i * getEntrySize() + _keyLen);

	CegoBTreeValue::Comparison valueComp1 = iv.comp(nv1, _pSchema);
	CegoBTreeValue::Comparison valueComp2;
	
	if ( i+1 == ne )
	{
	    valueComp2 = CegoBTreeValue::LESS;
	}
	else
	{
	    nv2.setPtr((char*)((long long)_pI + sizeof(int)) + (i+1)*getEntrySize(), _keyLen);
	    dp2.decode((char*)((long long)_pI + sizeof(int)) + (i+1)*getEntrySize() + _keyLen);

	    valueComp2 = iv.comp(nv2, _pSchema);
	}
		
	if ( valueComp1 == CegoBTreeValue::EQUAL || valueComp2 == CegoBTreeValue::EQUAL )
	{       	    
	    if ( valueComp1 == CegoBTreeValue::EQUAL )
	    {
		if ( isUnique && ! ( tid && checkForDeleted(pOM, dp1, tabSetId, tid) ) )
		{
		    Chain msg = Chain("Duplicate key ") + iv.toChain(_pSchema) + Chain(" on unique btree ") + btreeName;
		    throw Exception(EXLOC, msg);		    
		}
		else
		{
		    duplicateAdded = true;
		}
	    }
	    if (  valueComp2 == CegoBTreeValue::EQUAL )
	    {
		if ( isUnique && ! ( tid && checkForDeleted(pOM, dp2, tabSetId, tid) ) )
		{
		    Chain msg = Chain("Duplicate key ") + iv.toChain(_pSchema) + Chain(" on unique btree ") + btreeName;
		    throw Exception(EXLOC, msg);		    
		}
		else
		{
		    duplicateAdded = true;
		}
	    }
	}

	if ( ( valueComp1 == CegoBTreeValue::MORE || valueComp1 == CegoBTreeValue::EQUAL )
	     && ( valueComp2 == CegoBTreeValue::LESS || valueComp2 == CegoBTreeValue::EQUAL ) )
	{
	    // cout << "CASE-A" << endl;
	    i++;
	    posFound=true;
	}
	else if ( valueComp2 == CegoBTreeValue::MORE )
	{
	    // cout << "CASE-B" << endl;
	    if ( i == ne )
		posFound = true;
	    else
		lb=i;	    	    	    
	}
	else
	{
	    // cout << "CASE-C" << endl;
	    if ( i == 0 )
		posFound = true;
	    else
		rb=i;
	}
	// sleep(1);
    }

    // cout << "Position found at : " << i << endl;

    if ( i < ne )
    {
	shiftEntries(i, 1);
    }

    char* pKey = (char*)((long long)_pI + sizeof(int) + i* getEntrySize());    
    memcpy(pKey, iv.getPtr(), _keyLen);
    char* pDP = (char*)((long long)pKey + _keyLen);
    dp.encode(pDP);

    incEntries();

    return true;
}

bool CegoBTreeNode::deleteValue(const CegoBTreeValue& iv, const CegoDataPointer& dp)
{
    if ( _nt == NODE )
	throw Exception(EXLOC, Chain("Invalid method for node type"));

    int ne = numEntries();

    if ( ne == 0 )
	return false;

    char* pKey = (char*)((long long)_pI + sizeof(int));

    CegoBTreeValue lv;
    
    int i=0;
    while ( i < ne )
    {
	lv.setPtr(pKey, _keyLen);

	CegoDataPointer ldp;
	ldp.decode(pKey + _keyLen);
	
	if ( lv.isEqual(iv, _pSchema) && ldp == dp)
	{
	    shiftEntries(i+1, -1);	    
	    decEntries();
	    return true; 	    
	}
	
	pKey += getEntrySize();
	i++;
    }
    
    return false;
}

bool CegoBTreeNode::valueExists(const CegoBTreeValue& iv)
{
    CegoDataPointer dp;
    return valueExists(iv, dp);
}

bool CegoBTreeNode::valueExists(const CegoBTreeValue& iv, CegoDataPointer& dp)
{
    if ( _nt == NODE )
	throw Exception(EXLOC, Chain("Invalid method for node type"));

    int ne = numEntries();

    if ( ne == 0 )
	return false;

    char* pKey = (char*)((long long)_pI + sizeof(int));

    CegoBTreeValue lv;
    
    int i=0;
    while ( i < ne )
    {
	lv.setPtr(pKey, _keyLen);

	if ( lv.isEqual(iv, _pSchema))
	{
	    dp.decode(pKey + _keyLen);
	    return true; 	    
	}
	pKey += getEntrySize();
	i++;
    }

    return false;
}

bool CegoBTreeNode::addNode(const CegoBTreeValue& iv, const CegoBTreeNode& node)
{
    if ( _nt == LEAF )
	throw Exception(EXLOC, Chain("Invalid method for node type"));
    
    int ne = numEntries();       	
    if ( ne == _maxEntries || ne == 0 )
    {
	return false;
    }

    bool posFound=false;
    char* p = 0;
    int i=0;	    
    
    while ( i < ne && posFound == false)
    {
	
	p = (char*)((long long)_pI + sizeof(int) + sizeof(PageIdType) + i * getEntrySize());
	CegoBTreeValue nv(p, _keyLen);
	if ( nv.isHigher(iv, _pSchema) )
	{
	    shiftEntries(i, 1);
	    posFound=true;
	}
	else if ( iv.isEqual(nv, _pSchema) )
	{
	    // if equal, we also have to check the maximum value
	    if ( node.getMax().isHigher(nv, _pSchema) )
	    {
		i++;
	    }
	    else
	    {
		shiftEntries(i, 1);
		posFound=true;		    
	    }
	}
	else
	{	    
	    i++;
	}
    }
 
    ////////////////////////////////////
    // Setting up new node information //
    ////////////////////////////////////

    p = (char*)((long long)_pI + sizeof(int) + sizeof(PageIdType) + i * getEntrySize());
    
    // put index value
    memcpy(p, iv.getPtr(), _keyLen);
    p+=_keyLen;
    
    // put child pointer
    PageIdType pageId = node.getPageId();
    memcpy(p, &pageId, sizeof(PageIdType));
    p+=sizeof(PageIdType);
    
    incEntries();
    
    return true;
}

void CegoBTreeNode::split(CegoBTreeNode& n)
{

    int pos; 

    if ( _nt == LEAF )
    {
	pos = numEntries() / 2;
	char* srcPtr = (char*)((long long)_pI + sizeof(int) + pos * getEntrySize());
	char* destPtr = (char*)((long long)n._pI + sizeof(int));
	int s =  ( numEntries() - pos ) * getEntrySize();

	memcpy(destPtr, srcPtr, s);
	
    }
    else if ( _nt == NODE )
    {
	pos = numEntries() / 2;
	char* srcPtr = (char*)((long long)_pI + sizeof(int) + pos * getEntrySize());
	char* destPtr = (char*)((long long)n._pI + sizeof(int));
	int s =  ( numEntries() - pos ) * getEntrySize() + sizeof(PageIdType);

	memcpy(destPtr, srcPtr, s);	

    }

    int s1 = pos;
    int s2 = numEntries() - pos;
    memcpy(_pI, &s1, sizeof(int));
    memcpy(n._pI, &s2, sizeof(int));

}

void CegoBTreeNode::shiftEntries(int pos, int offset)
{
    if ( _nt == LEAF )
    {
	char* srcPtr = (char*)((long long)_pI + sizeof(int) + pos * getEntrySize());
	char* destPtr = (char*)((long long)_pI + sizeof(int) + ( pos + offset) * getEntrySize());
	int shiftSize = ( numEntries() - pos ) * getEntrySize();

	memmove(destPtr, srcPtr, shiftSize);

	/*
	if ( offset > 0 )
	{
	    while ( shiftSize )
	    {
		shiftSize--;
		*((char*)((long long)destPtr + (long long)shiftSize)) = *((char*)((long long)srcPtr + (long long)shiftSize));
	    }
	}
	else
	{
	    int i=0;
	    while ( i < shiftSize )
	    {
		*((char*)((long long)destPtr + (long long)i)) = *((char*)((long long)srcPtr + (long long)i));
		i++;
	    }
	}
	*/


    }
    else if ( _nt == NODE )
    {
	char* srcPtr = (char*)((long long)_pI + sizeof(int) + sizeof(PageIdType) + pos * getEntrySize());
	char* destPtr = (char*)((long long)_pI + sizeof(int) + sizeof(PageIdType) + ( pos + offset ) * getEntrySize());
	int shiftSize = ( numEntries() - pos ) * getEntrySize();


	memmove(destPtr, srcPtr, shiftSize);

	/*
	if ( offset > 0 )
	{
	    while ( shiftSize )
	    {
		shiftSize--;
		*((char*)((long long)destPtr + (long long)shiftSize)) = *((char*)((long long)srcPtr + (long long)shiftSize));
	    }
	}
	else
	{
	    int i=0;
	    while ( i < shiftSize )
	    {
		*((char*)((long long)destPtr + (long long)i)) = *((char*)((long long)srcPtr + (long long)i));
		i++;
	    }  
	}
	*/
    }
}

CegoBTreeValue CegoBTreeNode::getMin() const
{
    if ( numEntries() == 0 )
	throw Exception(EXLOC, Chain("No values values available in node"));
          
    if ( _nt == LEAF )    
    {
	char* p = (char*)((long long)_pI + sizeof(int));
	CegoBTreeValue iv(p, _keyLen);
	return iv;
    }
    else if ( _nt == NODE )
    {
	char* p = (char*)((long long)_pI + sizeof(int) + sizeof(PageIdType));
	CegoBTreeValue iv(p, _keyLen);
	return iv;
    }
    else
    {
	throw Exception(EXLOC, Chain("Invalid node type"));
    }
}

CegoBTreeValue CegoBTreeNode::getMax() const
{
    int pos = numEntries() - 1;

    if ( pos > 0 )
    {
	char* p;
	if ( _nt == LEAF )
	{
	    p = (char*)((long long)_pI + sizeof(int) + pos * getEntrySize());
	}
	else
	{
	    p = (char*)((long long)_pI + sizeof(int) + sizeof(PageIdType) + pos * getEntrySize());
	}
	CegoBTreeValue iv(p, _keyLen);
	return iv;

    }
    throw Exception(EXLOC, "No index values available in node");
}

bool CegoBTreeNode::propagate(const CegoBTreeValue& iv, const CegoBTreeNode& leftNode, const CegoBTreeNode& rightNode)
{
    if ( _nt == LEAF )
	throw Exception(EXLOC, Chain("Invalid method for node type"));
 
    int ne = numEntries();       	
    if ( ne == _maxEntries )
    {
	return false;
    }

    if ( ne == 0 )
    {
	// first entry
	
	char* p = (char*)((long long)_pI + sizeof(int));
	PageIdType pageId, nextPageId;
	
	// put left child pointer
	
	pageId = leftNode.getPageId();
	memcpy(p, &pageId, sizeof(PageIdType));
	p+=sizeof(PageIdType);

	// put index value
	memcpy(p, iv.getPtr(), _keyLen);
	p+=_keyLen;
	
	// put right child pointer
	nextPageId = rightNode.getPageId();
	memcpy(p, &nextPageId, sizeof(PageIdType));
	p+=sizeof(PageIdType);
	
	incEntries();
	
	return true;
    }
    else
    {

	bool posFound=false;
	char* p = 0;
	int i=0;	    

	while ( i < ne && posFound == false)
	{
	    p = (char*)((long long)_pI + sizeof(int) + sizeof(PageIdType) + i * getEntrySize());
	    CegoBTreeValue nv(p, _keyLen);

	    // cout << "Comparing " << iv.toChain(_pSchema) << " and " << nv.toChain(_pSchema) << "i=" << i << "ne=" << ne << endl;
	    
	    if ( iv.isHigher(nv, _pSchema) )
	    {		
		i++;
	    }
	    else if ( iv.isEqual(nv, _pSchema) )
	    {
		// if equal, we also have to check the maximum value
		if ( rightNode.getMax().isHigher(nv, _pSchema) )
		{
		    i++;
		}
		else
		{
		    shiftEntries(i, 1);
		    posFound=true;		    
		}
	    }
	    else
	    {
		shiftEntries(i, 1);
		posFound=true;
	    }
	}
	
	////////////////////////////////////
	// Setting up new node information //
	////////////////////////////////////

	p = (char*)((long long)_pI + sizeof(int) + sizeof(PageIdType) + i * getEntrySize());
	
	// put index value
	memcpy(p, iv.getPtr(), _keyLen);
	p+=_keyLen;
	
	// put child pointer
	PageIdType pageId = rightNode.getPageId();
	memcpy(p, &pageId, sizeof(PageIdType));
	p+=sizeof(PageIdType);
	
	incEntries();
	
	return true;

    }
}
 
CegoBTreeNode& CegoBTreeNode::operator = (const CegoBTreeNode& n)
{
    _nt = n._nt;
    _pI = n._pI;
    _len = n._len;
    _pageId = n._pageId;
    _nextPageId = n._nextPageId;
    _keyLen = n._keyLen;
    _maxEntries = n._maxEntries;
    _pSchema = n._pSchema;
    return (*this);
}

bool CegoBTreeNode::verify()
{
    int n = numEntries();

    if ( _nt == LEAF )
    {
	int i = 0;
	while ( i < n )
	{	    
	    char* srcPtr = (char*)((long long)_pI + sizeof(int) + i * getEntrySize());
	    CegoBTreeValue iv;
	    iv.setPtr(srcPtr, _keyLen);
	    srcPtr += _keyLen;
	    CegoDataPointer dp;
	    dp.decode(srcPtr);

	    // we check pageId and offset, must not be zero in any case
	    if ( dp.getPageId() == 0 || dp.getOffset() == 0 )
	    {
		return false;
	    }
	    // cout << iv.toChain(_pSchema) << " " << dp << endl;	  
	    i++;
	}
    }
    else if ( _nt == NODE )
    {
	PageIdType pageId;

	char* p = (char*)((long long)_pI + sizeof(int));
	
	memcpy(&pageId, p, sizeof(PageIdType));
			  
	// cout << levelIndent(level) << "Child : " << fileId << "," << pageId << endl;

	// we check fileId, must not be zero in any case
	if ( pageId == 0  )
	{
	    return false;
	}
	
	int i = 0;
	while ( i < n )
	{
	    // cout << levelIndent(level) << "Key " << i << " : ";
	    
	    p = (char*)((long long)_pI + sizeof(int) + sizeof(PageIdType) + i * getEntrySize());
	    CegoBTreeValue iv;
	    iv.setPtr(p, _keyLen);
	    // cout << iv.toChain(_pSchema) << endl;
	    p += _keyLen;
	    
	    memcpy(&pageId, p, sizeof(PageIdType));

	    // we check fileId, must not be zero in any case
	    if ( pageId == 0 )
	    {
		return false;
	    }

	    // cout << levelIndent(level) << "Child : " << fileId << "," << pageId << endl;

	    i++;
	}
	// cout << levelIndent(level) << "################## " << endl;
    }
    else
    {
	return false;
    }

    return true;
}

void CegoBTreeNode::printNode(int level)
{
    int n = numEntries();

    if ( _nt == LEAF )
    {

	cout << levelIndent(level) << "------- LEAF -------" << endl;
	cout << levelIndent(level) << "NumEntries=" << n << " PageId=" << _pageId << endl;

	int i = 0;
	while ( i < n )
	{
	    cout << levelIndent(level) << "Entry " << i << " : ";
	    
	    char* srcPtr = (char*)((long long)_pI + sizeof(int) + i * getEntrySize());
	    CegoBTreeValue iv;
	    iv.setPtr(srcPtr, _keyLen);
	    srcPtr += _keyLen;
	    CegoDataPointer dp;
	    dp.decode(srcPtr);
	    cout << iv.toChain(_pSchema) << " " << dp << endl;	  
	    i++;
	}
	cout << levelIndent(level) << "Next Page : " << _nextPageId << endl;
	cout << levelIndent(level) << "--------------------" << endl;
    }
    else if ( _nt == NODE )
    {
	cout << levelIndent(level) << "####### NODE ####### " << endl;
	cout << levelIndent(level) << "NumEntries=" << n << " PageId=" << _pageId << endl;

	PageIdType pageId;

	char* p = (char*)((long long)_pI + sizeof(int));
	
	memcpy(&pageId, p, sizeof(PageIdType));
			  
	cout << levelIndent(level) << "Child : " << pageId << endl;

	int i = 0;
	while ( i < n )
	{
	    cout << levelIndent(level) << "Key " << i << " : ";
	    
	    p = (char*)((long long)_pI + sizeof(int) + sizeof(PageIdType) + i * getEntrySize());
	    CegoBTreeValue iv;
	    iv.setPtr(p, _keyLen);
	    cout << iv.toChain(_pSchema) << endl;
	    p += _keyLen;
	    
	    memcpy(&pageId, p, sizeof(PageIdType));
	    
	    cout << levelIndent(level) << "Child : " << pageId << endl;


	    i++;
	}
	cout << levelIndent(level) << "################## " << endl;
    }
    else
    {
	cout << levelIndent(level) << "??? UNDEFINED ??? " << endl;
    }
}

void CegoBTreeNode::reset()
{
    _pNextChild = (char*)((long long)_pI + sizeof(int));
}

bool CegoBTreeNode::nextChildPointer(PageIdType &pageId)
{
    if ( _nt == LEAF )
	throw Exception(EXLOC, Chain("Invalid method for node type"));

    if ( (long long)_pNextChild > (long long)_pI + sizeof(int) + numEntries() * getEntrySize() )
	return false;
    
    memcpy(&pageId, _pNextChild, sizeof(PageIdType));
    _pNextChild+=sizeof(PageIdType);
    
    _pNextChild+=_keyLen;
    
    return true;
}

int CegoBTreeNode::getMedPage(int lb, int rb, PageIdType& pageId)
{
    if ( _nt == LEAF )
	throw Exception(EXLOC, Chain("Invalid method for node type"));
    
    int med = (lb + rb) / 2;
    char* _pMed = (char*)((long long)_pI + sizeof(int) + med * getEntrySize());
    
    memcpy(&pageId, _pMed, sizeof(PageIdType));
   
    return med;
}

bool CegoBTreeNode::rightValue(int pos, CegoBTreeValue& val)
{
    if ( _nt == LEAF )
	throw Exception(EXLOC, Chain("Invalid method for node type"));

    char* pChild = (char*)((long long)_pI + sizeof(int) + pos * getEntrySize());
    if ( (long long)pChild > (long long)_pI + sizeof(int) + numEntries() * getEntrySize() )
	return false;
    
    
    pChild+=sizeof(PageIdType);

    
    val.setPtr(pChild, _keyLen);
    return true;
}


void CegoBTreeNode::getLastChildPointer(PageIdType &pageId) const
{
    if ( _nt == LEAF )
	throw Exception(EXLOC, Chain("Invalid method for node type"));

    if ( numEntries() == 0 )
    {
	pageId = 0;
    }
    else
    {
	char *p = (char*)((long long)_pI + sizeof(int) + numEntries() * getEntrySize());
    
	memcpy(&pageId, p, sizeof(PageIdType));
    }
}

void CegoBTreeNode::setFirstChildPointer(PageIdType pageId)
{
    if ( _nt == LEAF )
	throw Exception(EXLOC, Chain("Invalid method for node type"));

    char *p = (char*)((long long)_pI + sizeof(int));
    
    memcpy(p, &pageId, sizeof(PageIdType));
}

bool CegoBTreeNode::rightChild(CegoBTreeValue& val)
{
    if ( _nt == LEAF )
	throw Exception(EXLOC, Chain("Invalid method for node type"));
   
    if ( (long long)_pNextChild > (long long)_pI + sizeof(int) + numEntries() * getEntrySize() )
	return false;
    
    val.setPtr(_pNextChild - _keyLen, _keyLen);
    return true;
}

bool CegoBTreeNode::nextValue(CegoBTreeValue& val, CegoDataPointer& dp)
{
    if ( _nt == NODE )
	throw Exception(EXLOC, Chain("Invalid method for node type"));

    if ( (long long)_pNextChild >= (long long)_pI + sizeof(int) + numEntries() * getEntrySize() )
	return false;
    
    val.setPtr(_pNextChild, _keyLen);
    _pNextChild +=_keyLen;
    dp.decode(_pNextChild);
    _pNextChild += DPENCODINGLENGTH;
    
    return true;    
}

Chain CegoBTreeNode::levelIndent(int level)
{
    Chain s;
    while ( level > 0 )
    {
	s += Chain(" ");
	level--;
    }
    return s;
}

bool CegoBTreeNode::checkForDeleted(CegoObjectManager* pOM, const CegoDataPointer &dp, int tabSetId, unsigned long long tid)
{    
    char* p;
    int len;

    CegoBufferPage bp = pOM->claimDataPtrUnlocked(tabSetId, CegoBufferPool::SYNC, dp, p, len);
    
    // skipping tid
    unsigned long long dataTid;
    unsigned long long tastep;
    CegoTupleState ts;

    CegoQueryHelper::decodeTupleHeader(dataTid, tastep, ts, p);
    
    pOM->releaseDataPtrUnlocked(bp, false);

    /*
    cout << "Datapointer = " << dp << endl;
    cout << "tid = " << tid << endl;
    cout << "dataTid = " << dataTid << endl;
    cout << "ts = " << ts << endl;
    */

    if (tid == dataTid && ( ts == DELETED || ts == OBSOLETE ))
    {
	return true;
    }
    else
    {
	return false;
    }
}
